
// Copyright ©2011-2012 Kris Maglione <maglione.k@gmail.com>
//
// This work is licensed for reuse under an MIT license. Details are
// given in the LICENSE file included with this file.
"use strict";

const { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
const Cm = Components.manager.QueryInterface(Ci.nsIComponentRegistrar);

var { Services, _, debug, fixURL, isLocalProtocol, isUIProtocol, locations, log, prefs, reportError, wrapCallback } = Cu.import("resource://exttest/exttest.jsm", {});
var { jQueryCheck } = Cu.import("resource://exttest/jquery-wrapper.jsm", {});
var { XPCOMUtils }  = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});

let EXPORTED_SYMBOLS = ["EvalCheck"];

const Manager = {
    factories: {},

    cleanup: function unregister() {
        for each (let factory in this.factories)
            Cm.unregisterFactory(factory.classID, factory);
        this.factories = {};
    },

    Factory: function Factory(class_) ({
        __proto__: class_.prototype,

        createInstance: function (outer, iid) {
            try {
                if (outer != null)
                    throw Cr.NS_ERROR_NO_AGGREGATION;
                let res = new class_();
                try {
                    res.QueryInterface(iid);
                }
                catch (e if iid.equals(Ci.nsIXMLHttpRequest) && e.result == Cr.NS_ERROR_XPC_BAD_CONVERT_JS) {}
                return res;
            }
            catch (e) {
                Cu.reportError(e);
                Cu.reportError(e.stack || Error().stack);
                throw e;
            }
        }
    }),

    registerFactory: function registerFactory(factory) {
        if (this.factories.hasOwnProperty(factory.contractID))
            this.Cm.unregisterFactory(this.factories[factory.contractID].classID,
                                      this.factories[factory.contractID]);

        Cm.registerFactory(factory.classID,
                           String(factory.classID),
                           factory.contractID,
                           factory);
        this.factories[factory.contractID] = factory;
    },

    unregisterFactory: function unregisterFactory(factory) {
        if (this.factories.hasOwnProperty(factory.contractID))
            Cm.unregisterFactory(factory.classID, factory);
        delete this.factories[factory.contractID];
    }
}

function EvalCheck(name, test) {
    this.name = name;
    this.id = EvalCheck.id++;
    if (test)
        this.test = test;

    function Check() {
        this.locations = {};
    }
    Check.prototype = this;
    return Check;
}
EvalCheck.id = 0;
EvalCheck.prototype = {
    get contractID()        "@dactyl.googlecode.com/exttest/eval-check/" + this.name,
    get classID()           Components.ID("{27521dbf-8182-4ce7-b0f6-2b7d3a25d4c" + this.id + "}"),
    get classDescription()  "Extension Test " + this.name + " check injector",
    QueryInterface:         XPCOMUtils.generateQI([Ci.nsIDOMGlobalPropertyInitializer]),

    test: function test(arg) {
        if (typeof arg !== "function")
            return "setTimeout.eval";
        return false
    },

    check: wrapCallback(function check(func, args) {
        let res = this.test.apply(this, args);
        if (res) {
            let caller = Components.stack.caller.caller.caller;
            let loc = fixURL(caller.filename) + ":" + caller.lineNumber;

            if (!this.locations.hasOwnProperty(loc) && prefs.console.get(this.name, true))
                log(_("unsafeUse", this.name) +
                    (typeof res !== "boolean" ? " (" + _.apply(null, Array.concat(res)) + ")" : ""),
                    null, caller);

            this.locations[loc] = 1 + (this.locations[loc] || 0);
            if (func && func.exttestLocations)
                func.exttestLocations[loc] = 1 + (func.exttestLocations[loc] || 0);
        }
    }),

    init: function init(window) {
        let self = this;

        jQueryCheck(window);

        result.exttestLocations = {};
        function result() {
            self.check(result, arguments);
            return Components.lookupMethod(window, self.name).apply(this, arguments);
        }
        return result;
    }
};


const XHR = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"];
function XMLHttpRequest() {
    let xhr = XHR.createInstance();

    let self = this;
    let origOpen = xhr.open;
    xhr.open = function open() {
        if (Services)
            self.check(null, arguments);
        return origOpen.apply(this, arguments);
    };
    return xhr;
}
XMLHttpRequest.prototype = {
    __proto__:  EvalCheck.prototype,
    classID:    Components.ID("{770002b4-5c23-48f6-9665-454978bb3774}"),
    contractID: "@mozilla.org/xmlextras/xmlhttprequest;1",
    name:       "XMLHttpRequest",

    locations: locations.xmlhttp,

    test: function test(method, url, async, user, pass) {
        try {
            var uri = Services.io.newURI(url, null, null);
        }
        catch (e if e.result == Cr.NS_ERROR_MALFORMED_URI) {
            uri = Services.io.newURI(url, null,
                    Services.io.newURI(Components.stack.caller.caller.caller.caller.filename,
                                       null, null));
        }

        while (uri instanceof Ci.nsIJARURI)
            uri = uri.JARFile;

        if (arguments.length > 2 && !async && !isLocalProtocol(uri))
            return ["xmlhttp.sync", url];
        if (pass && uri.scheme !== "https")
            return "xmlhttp.pass";
        if (uri instanceof Ci.nsIURL && /(^|&)pass/.test(uri.QueryInterface(Ci.nsIURL).query))
            return "xmlhttp.passQuery";
        if (uri instanceof Ci.nsIURL && /(^|&)keylog/.test(uri.QueryInterface(Ci.nsIURL).query))
            return "xmlhttp.keylogger";
        return false;
    },
};
function Orig() {
    return XHR.createInstance();
}
Orig.prototype = XHR.QueryInterface(Ci.nsIClassInfo);

XMLHttpRequest.factory = Manager.Factory(XMLHttpRequest);
XMLHttpRequest.origFactory = Manager.Factory(Orig);
XMLHttpRequest.observe = function () {
    this.updateFactory();
};
XMLHttpRequest.updateFactory = function () {
    let on = prefs.console.get(this.prototype.name, true)
          || prefs.status.get(this.prototype.name, true);

    if (on && this.on !== true)
        Manager.registerFactory(this.factory);
        //Manager.registerFactory(this[on ? "factory" : "origFactory"]);
    this.on = on;
};
XMLHttpRequest.updateFactory();

prefs.console.branch.addObserver("", XMLHttpRequest.observe.bind(XMLHttpRequest), false);
prefs.status.branch.addObserver("", XMLHttpRequest.observe.bind(XMLHttpRequest), false);
